//
//  PithosAccount.h
//  pithos-macos
//
// Copyright 2012 GRNET S.A. All rights reserved.
//
// Redistribution and use in source and binary forms, with or
// without modification, are permitted provided that the following
// conditions are met:
// 
//   1. Redistributions of source code must retain the above
//      copyright notice, this list of conditions and the following
//      disclaimer.
// 
//   2. Redistributions in binary form must reproduce the above
//      copyright notice, this list of conditions and the following
//      disclaimer in the documentation and/or other materials
//      provided with the distribution.
// 
// THIS SOFTWARE IS PROVIDED BY GRNET S.A. ``AS IS'' AND ANY EXPRESS
// OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GRNET S.A OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
// USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
// AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
// ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
// 
// The views and conclusions contained in the software and
// documentation are those of the authors and should not be
// interpreted as representing official policies, either expressed
// or implied, of GRNET S.A.

#import <Foundation/Foundation.h>
@class PithosSyncDaemon;
@class ASIPithos;
@class ASIPithosRequest;
@class PithosAccountNode;
@class PithosSharingAccountsNode;

@interface PithosAccount : NSObject <NSCoding> {
    NSString *uniqueName;
    BOOL active;
    NSString *name;
    NSString *clientVersion;
    
    BOOL syncActive;
    NSString *syncDirectoryPath;
    NSMutableDictionary *syncAccountsDictionary;
    BOOL syncSkipHidden;
    NSDate *syncLastCompleted;
    PithosSyncDaemon *syncDaemon;
        
    // Base for all necessary URLs, default "https://pithos.okeanos.grnet.gr"
    NSString *serverURL;
    // Appended to serverURL to get storageURLPrefix and authURL, default "v1"
    // Can be overriden by setting directly storageURLPrefix and authURL
    NSString *versionResource;
    // Appended to serverURL to get loginURLPrefix, default "login"
    // Can be overriden by setting directly loginURLPrefix
    NSString *loginResource;
    // Appended to serverURL to get publicURLPrefix, default nil
    // Can be overriden by setting directly publicURLPrefix
    NSString *publicResource;
    // Appended to serverURL to get userCatalogURL, default "user_catalogs"
    // Can be overriden by setting directly userCatalogURL
    NSString *userCatalogResource;

    NSString *authUser;
    NSString *authToken;
    NSString *storageURLPrefix;
    NSString *authURL;
    NSString *publicURLPrefix;
    NSString *loginURLPrefix;
    NSString *userCatalogURL;
    NSMutableDictionary *userCatalog;
    
    ASIPithos *pithos;
    PithosAccountNode *accountNode;
    PithosSharingAccountsNode *sharingAccountsNode;
    
    BOOL updatePithos;
    BOOL resetSyncDaemonLocalState;
}

@property (nonatomic, strong) NSString *uniqueName;
@property (assign) BOOL active;
@property (nonatomic, strong) NSString *name;
@property (nonatomic, strong) NSString *clientVersion;

@property (nonatomic, assign) BOOL syncActive;
@property (nonatomic, strong) NSString *syncDirectoryPath;
@property (nonatomic, strong) NSMutableDictionary *syncAccountsDictionary;
@property (nonatomic, assign) BOOL syncSkipHidden;
@property (nonatomic, strong) NSDate *syncLastCompleted;
@property (nonatomic, strong) PithosSyncDaemon *syncDaemon;

@property (nonatomic, strong) NSString *serverURL;
@property (nonatomic, strong) NSString *versionResource;
@property (nonatomic, strong) NSString *loginResource;
@property (strong) NSString *publicResource;
@property (nonatomic, strong) NSString *userCatalogResource;

@property (nonatomic, strong) NSString *authUser;
@property (nonatomic, strong) NSString *authToken;
@property (nonatomic, strong) NSString *storageURLPrefix;
@property (nonatomic, strong) NSString *authURL;
@property (nonatomic, strong) NSString *publicURLPrefix;
@property (nonatomic, strong) NSString *loginURLPrefix;
@property (nonatomic, strong) NSString *userCatalogURL;
@property (nonatomic, strong) NSMutableDictionary *userCatalog;

@property (strong, readonly) ASIPithos *pithos;
@property (readonly) PithosAccountNode *accountNode;
@property (readonly) PithosSharingAccountsNode *sharingAccountsNode;

+ (id)pithosAccount;

- (BOOL)urlIsValid:(NSString *)urlString;
- (void)authenticateWithServerURL:(NSString *)aServerURL authUser:(NSString *)anAuthUser authToken:(NSString *)anAuthToken;
- (void)loginWithServerURL:(NSString *)aServerURL;
- (void)updateSyncWithSyncActive:(BOOL)aSyncActive 
               syncDirectoryPath:(NSString *)aSyncDirectoryPath 
          syncAccountsDictionary:(NSMutableDictionary *)aSyncAccountsDictionary 
                  syncSkipHidden:(BOOL)aSyncSkipHidden;
- (ASIPithosRequest *)updateUserCatalogForForDisplaynames:(NSArray *)displaynames UUIDs:(NSArray *)UUIDs;
- (NSString *)displaynameForUUID:(NSString *)UUID safe:(BOOL)safe;
- (NSString *)displaynameForUUID:(NSString *)UUID;

@end
