//
//  PithosSyncDaemon.h
//  pithos-macos
//
// Copyright 2011-2012 GRNET S.A. All rights reserved.
//
// Redistribution and use in source and binary forms, with or
// without modification, are permitted provided that the following
// conditions are met:
// 
//   1. Redistributions of source code must retain the above
//      copyright notice, this list of conditions and the following
//      disclaimer.
// 
//   2. Redistributions in binary form must reproduce the above
//      copyright notice, this list of conditions and the following
//      disclaimer in the documentation and/or other materials
//      provided with the distribution.
// 
// THIS SOFTWARE IS PROVIDED BY GRNET S.A. ``AS IS'' AND ANY EXPRESS
// OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GRNET S.A OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
// USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
// AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
// ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
// 
// The views and conclusions contained in the software and
// documentation are those of the authors and should not be
// interpreted as representing official policies, either expressed
// or implied, of GRNET S.A.

#import <Foundation/Foundation.h>
@class PithosAccount;
@class ASIPithos;
@class PithosActivityFacility;
@class ASINetworkQueue;

@interface PithosSyncDaemon : NSObject {
    NSString *directoryPath;
    PithosAccount *pithosAccount;
    NSDictionary *accountsDictionary;
    BOOL skipHidden;
    ASIPithos *pithos;

    NSMutableDictionary *userCatalog;
    NSUInteger accountsCount;
    NSMutableArray *accountsNames;
    NSUInteger accountsIndex;
    NSMutableDictionary *accountsPithosContainers;
    NSUInteger containersIndex;
    
    NSMutableArray *objects;
    NSMutableDictionary *remoteObjects;
    NSMutableDictionary *previousRemoteObjects;
    NSMutableDictionary *storedLocalObjectStates;
    NSMutableDictionary *currentLocalObjectStates;
    
    NSString *userCatalogFilePath;
    NSString *pithosStateFilePath;
    NSString *tempDownloadsDirPath;
    NSString *tempTrashDirPath;
    
    BOOL daemonActive;
    
    NSUInteger syncOperationCount;
    BOOL newSyncRequested;
    NSDate *lastCompletedSync;
    BOOL syncIncomplete;
    BOOL syncLate;
    
    ASINetworkQueue *networkQueue;
    NSOperationQueue *callbackQueue;
    
    PithosActivityFacility *activityFacility;
}

@property (nonatomic, strong) NSString *directoryPath;
@property (nonatomic, strong) NSDictionary *accountsDictionary;
@property (nonatomic, assign) BOOL skipHidden;
@property (nonatomic, strong) ASIPithos *pithos;

@property (nonatomic, strong) NSMutableDictionary *userCatalog;
@property (nonatomic, strong) NSMutableArray *accountsNames;
@property (nonatomic, strong) NSMutableDictionary *accountsPithosContainers;

@property (nonatomic, strong) NSMutableDictionary *remoteObjects;
@property (nonatomic, strong) NSMutableDictionary *previousRemoteObjects;
@property (nonatomic, strong) NSMutableDictionary *storedLocalObjectStates;
@property (nonatomic, strong) NSMutableDictionary *currentLocalObjectStates;

@property (nonatomic, strong, readonly) NSString *userCatalogFilePath;
@property (nonatomic, strong, readonly) NSString *pithosStateFilePath;
@property (nonatomic, strong, readonly) NSString *tempDownloadsDirPath;
@property (nonatomic, strong, readonly) NSString *tempTrashDirPath;

@property (nonatomic, strong) NSDate *lastCompletedSync;

// accountsDictionary contains entries of the form
// { <included sharing account name> : { <included container name> : [ <excluded directory names> ] } }
// <included sharing account name> :
//     the empty string "" denotes the user's own account
// <excluded directory names> : 
//     the empty string "" denotes all file objects in the container root
//     non-empty strings denote also container root file objects with these names
// Case insensitive filesystem is assumed, so directory names should be in lowercase
// and any trailing '/' should be replaced by ':'
- (id)initWithDirectoryPath:(NSString *)aDirectoryPath 
              pithosAccount:(PithosAccount *)aPithosAccount 
         accountsDictionary:(NSDictionary *)anAccountsDictionary 
                 skipHidden:(BOOL)aSkipHidden 
            resetLocalState:(BOOL)resetLocalState;

- (void)resetDaemon;
- (void)startDaemon;

- (void)sync;
- (BOOL)isSyncing;
- (void)syncLate;

@end
